/*------------------------------------------------------------------------------
* preceph.c : input precise ephemeris sp3 file
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2009/01/18 1.0 new
*           2009/01/31 1.1 fix bug on numerical error to read sp3a ephemeris
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

#define NMAX        10      /* order of polynomial interpolation */
#define MAXDTE      900.0   /* max time difference to ephemeris time (s) */

/* read sp3 header -----------------------------------------------------------*/
static int readsp3h(FILE *fp, gtime_t *time, char *type, int *sats,
                    double *bfact, char *tsys)
{
    int i,j,ns=0,sat;
    char buff[1024];
    
    trace(3,"readsp3h:\n");
    
    for (i=0;i<22;i++) {
        if (!fgets(buff,sizeof(buff),fp)) break;
        
        if (i==0) {
            *type=buff[2];
            if (str2time(buff,3,28,time)) return 0;
        }
        else if (2<=i&&i<=6) {
            for (j=0;j<17;j++) {
                sat=SAT((int)str2num(buff,10+3*j,2));
                if (1<=sat&&sat<=MAXSAT) sats[ns++]=sat;
            }
        }
        else if (i==12) {
            strncpy(tsys,buff+9,3); tsys[3]='\0';
        }
        else if (i==14) {
            bfact[0]=str2num(buff,3,10);
            bfact[1]=str2num(buff,14,12);
        }
    }
    return ns;
}
/* read sp3 body -------------------------------------------------------------*/
static int readsp3b(FILE *fp, char type, int *sats, int ns, double *bfact,
                    char *tsys, peph_t **peph)
{
    double std,ep[6],val,fact[]={1E3,1E3,1E3,1E-6};
    int i,j,sat,neph=0,nmax=0;
    char buff[1024];
    peph_t peph0={{0}},*p;
    
    trace(3,"readsp3b:\n");
    
    *peph=NULL;
    
    while (fgets(buff,sizeof(buff),fp)!=NULL) {
        if (!strncmp(buff,"EOF",3)) break;
        
        if (sscanf(buff+2,"%lf %lf %lf %lf %lf %lf",ep,ep+1,ep+2,ep+3,ep+4,ep+5)<6) {
            continue;
        }
        if (nmax==0||neph>=nmax) {
            nmax+=1024;
            if (!(*peph=(peph_t *)realloc(*peph,sizeof(peph_t)*nmax))) return 0;
        }
        p=*peph+neph;
        *p=peph0;
        p->time=epoch2time(ep);
        if (!strcmp(tsys,"UTC")) p->time=utc2gpst(p->time);
        
        for (i=0;i<ns*(type=='P'?1:2);i++) {
            if (!fgets(buff,sizeof(buff),fp)) break;
            if (buff[0]!='P'||(buff[1]!=' '&&buff[1]!='G')) continue;
            if ((sat=SAT((int)str2num(buff,2,2)))>MAXSAT) continue;
            
            for (j=0;j<4;j++) {
                if ((val=str2num(buff,j*14+4,14))==0.0) continue;
                if (fabs(val-999999.999999)<1E-6) continue;
                p->pos[sat-1][j]=val*fact[j];
                if (j<3) {
                    std=str2num(buff,j*3+61,2);
                    p->std[sat-1][j]=bfact[0]<=0.0?0.0:pow(bfact[0],std)*1E-3;
                }
                else {
                    std=str2num(buff,j*3+61,3);
                    p->std[sat-1][j]=bfact[1]<=0.0?0.0:pow(bfact[1],std)*1E-12;
                }
            }
        }
        neph++;
    }
    return neph;
}
/* read sp3 file -------------------------------------------------------------*/
static int readsp3fp(FILE *fp, peph_t **peph)
{
    gtime_t time;
    double bfact[2];
    int ns,sats[MAXSAT];
    char type,tsys[4]="";
    
    trace(3,"readsp3fp:\n");
    
    if ((ns=readsp3h(fp,&time,&type,sats,bfact,tsys))<=0) return 0;
    
    return readsp3b(fp,type,sats,ns,bfact,tsys,peph);
}
/* read satellite antenna phase center position --------------------------------
* read satellite antenna phase center position
* args   : char   *file       I   antenna pcv file (antex)
*          double *ppos       O   sat antenna phase center pos (m)
* return : status (1:ok,0:out of range/satellite number error)
*-----------------------------------------------------------------------------*/
extern int readsap(const char *file, gtime_t time, double sap[][3])
{
    FILE *fp;
    const double posp[2][3]={{0.279,0.0,1.023},{0.0,0.0,0.0}};
    const double epswp[]={2006,11,5,0,0,0}; /* igs antenna model swap */
    gtime_t ts={0},te={0},t0={0},tswp;
    double pos[3];
    char buff[1024];
    int i,j,sat=-1,block=0;
    
    trace(3,"readsap: file=%s\n",file);
    
    if (!(fp=fopen(file,"r"))) {
        trace(1,"readsap: file open error %s\n",file);
        return 0;
    }
    tswp=epoch2time(epswp);
    for (i=0;i<MAXSAT;i++) for (j=0;j<3;j++) sap[i][j]=0.0;
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (strlen(buff)<60) continue;
        if (strstr(buff+60,"END OF ANTENNA")) {
            sat=-1; ts=t0; te=t0;
        }
        if (strstr(buff+60,"START OF ANTENNA")) sat=0;
        if (sat<0) continue;
        if (strstr(buff+60,"TYPE / SERIAL NO")) {
            if (buff[20]=='G') sat=(int)str2num(buff,21,2);
            if (strstr(buff,"BLOCK IIA")==buff) block=0; else block=1;
        }
        else if (strstr(buff+60,"VALID FROM")) {
            if (!str2time(buff,0,43,&ts)) continue;
        }
        else if (strstr(buff+60,"VALID UNTIL")) {
            if (!str2time(buff,0,43,&te)) continue;
        }
        else if (strstr(buff+60,"NORTH / EAST / UP")) {
            if (sscanf(buff,"%lf %lf %lf",pos,pos+1,pos+2)<3) continue;
            if (ts.time!=0&&timediff(ts,time)>0.0) continue;
            if (te.time!=0&&timediff(te,time)<0.0) continue;
            if (sat<1||MAXSAT<sat) continue;
            if (timediff(time,tswp)<0.0) {
                for (i=0;i<3;i++) sap[SAT(sat)-1][i]=posp[block][i];
            }
            else {
                for (i=0;i<3;i++) sap[SAT(sat)-1][i]=pos[i]*1E-3;
            }
        }
    }
    fclose(fp);
    return 1;
}
/* read sp3 file ---------------------------------------------------------------
* read sp3 precise ephmeris file
* args   : char   *file       I   sp3 precise ephemeris file
*          peph_t **peph      IO  precise ephemeris
* return : number of precise epheremeris (0:error)
*-----------------------------------------------------------------------------*/
extern int readsp3(const char *file, peph_t **peph)
{
    FILE *fp;
    int n;
    
    trace(3,"readsp3: file=%s\n",file);
    
    if (!(fp=fopen(file,"r"))) {
        trace(1,"readsp3: file open error: %s\n",file);
        return 0;
    }
    n=readsp3fp(fp,peph);
    fclose(fp);
    return n;
}
/* polynomial interpolation by neville's algorithm ---------------------------*/
static double interppol(const double *x, double *y, int n)
{
    int i,j;
    
    for (j=1;j<n;j++) {
        for (i=0;i<n-j;i++) {
            y[i]=(x[i+j]*y[i]-x[i]*y[i+1])/(x[i+j]-x[i]);
        }
    }
    return y[0];
}
/* satellite precise ephemeris to satellite position/clock-bias ----------------
* compute satellite position with precise ephemeris
* args   : gtime_t time       I   time
*          int     sat        I   satellite number
*          peph_t *eph        I   precise ephemeris records (sorted by time)
*          int    neph        I   number of precise ephemeris
*          double *rs         O   satellite position (ecef) (m)
*          double *dts        O   satellite clock-bias (s)
* return : status (1:ok,0:error or data outage)
*-----------------------------------------------------------------------------*/
extern int eph2posp(gtime_t time, int sat, peph_t *eph, int neph, double *rs,
                    double *dts)
{
    double x[NMAX+1],y[NMAX+1];
    int i,j,k,n,m;
    
    rs[0]=rs[1]=rs[2]=*dts=0.0;
    
    if (neph<NMAX+1||sat<1||MAXSAT<sat) return 0;
    
    if (timediff(time,eph[0].time)<-MAXDTE||
        timediff(time,eph[neph-1].time)>MAXDTE) return 0;
    
    /* binary search */
    for (i=0,j=neph-1;i<j;) {
        k=(i+j)/2;
        if (timediff(eph[k].time,time)<0.0) i=k+1; else j=k;
    }
    n=i; m=n-(NMAX+1)/2;
    if (m<0) m=0; else if (m+NMAX>=neph) m=neph-NMAX-1;
    
    for (i=0;i<=NMAX;i++) {
        x[i]=timediff(eph[m+i].time,time);
        if (norm(eph[m+i].pos[sat-1],3)<=0.0) return 0; /* orbit outage */
    }
    /* polynomial interpolation for orbit */
    for (i=0;i<3;i++) {
        for (j=0;j<=NMAX;j++) y[j]=eph[m+j].pos[sat-1][i];
        rs[i]=interppol(x,y,NMAX+1);
    }
    /* bilinear interpolation for clock */
    if (n<=0) {
        *dts=eph[0].pos[sat-1][3];
        return *dts!=0.0;
    }
    for (j=0;j<2;j++) {
        x[j]=timediff(eph[n+j-1].time,time);
        if ((y[j]=eph[n+j-1].pos[sat-1][3])==0.0) return 0;
    }
    *dts=interppol(x,y,2);
    return 1;
}
/* satellite positions/clock-biases with precise ephemeris ---------------------
* compute satellite positions and clock-biases with precise ephemeris
* args   : gtime_t time     I   time to search ephemeris
*          obsd_t *obs      I   observation data records
*          int     n        I   number of observation data records
*          nav_t  *nav      I   navigation messages
*          double *rs       O   satellilte positions (ecef) (m)
*          double *dts      O   satellilte clock-bias (sec)
* return : none
* notes  : rs[(0:2)+i*3],dts[i] = obs[i] satellite position/clock-bias
*          if no navigation messages, set 0.0 to rs[], dts[]
*-----------------------------------------------------------------------------*/
extern void satposp(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                    double *rs, double *dts)
{
    double rsp[3],rss[3],vs[3],tt,r;
    int i,j,sat;
    
    for (i=0;i<n&&i<MAXOBS;i++) {
        rs[i*3]=rs[1+i*3]=rs[2+i*3]=dts[i]=0.0;
        if (obs[i].P[0]==0.0) continue;
        sat=obs[i].sat;
        time=timeadd(obs[i].time,-obs[i].P[0]/CLIGHT);
        if (!eph2posp(time,sat,nav->peph,nav->ne,rsp,&tt)) continue;
        time=timeadd(time,-tt);
        if (!eph2posp(time,sat,nav->peph,nav->ne,rss,dts+i)) continue;
        r=norm(rss,3);
        for (j=0;j<3;j++) {
            rs[j+i*3]=rss[j]-nav->sap[sat-1][2]*rss[j]/r; /* ant z-pos */
            vs[j]=-(rss[j]-rsp[j])/tt;
        }
        dts[i]-=2.0*dot(rs+i*3,vs,3)/CLIGHT/CLIGHT; /* relativ */
    }
}
